from datetime import datetime
from io import TextIOWrapper


def write_rdf(
    f: TextIOWrapper,
    rxn_blocks: list[str],
    rxn_ids: list[str] = None,
):
    """Write a RDF file from a list of reaction blocks.

    Parameters
    ----------
    f : TextIOWrapper
        The file to write to.
    rxn_blocks : list[str]
        The reaction blocks to write.
    rxn_ids : list[str], optional
        The reaction IDs to use. Defaults to None. If None, sequential 5 digit
        numbers are used.
    """
    # current date and time as DD/MM/YY HH:MM
    datm = datetime.now().strftime("%d/%m/%y %H:%M")

    rdf_header = f"$RDFILE 1\n$DATM {datm}\n"
    rxn_header = "$RFMT $RIREG {}\n"

    if rxn_ids is None:
        # generate sequential 5 digit numbers
        rxn_ids = [f"{i:05d}" for i in range(1, len(rxn_blocks) + 1)]

    f.write(rdf_header)

    for rxn_id, rxn_block in zip(rxn_ids, rxn_blocks):
        f.write(rxn_header.format(rxn_id))
        f.write(rxn_block)



from typing import Any

from rdkit.Chem import Mol, MolFromMolBlock, MolFromSmiles, MolToMolBlock, MolToSmiles

from rdfreader.exceptions import InvalidMoleculeError
from rdfreader.parse.molblock import get_mol_block_metadata


class Molecule:
    def __init__(
        self,
        mol_block: str = None,
        properties: dict[str, Any] = dict(),
        except_on_invalid_molecule: bool = True,
        component_type: str = None,
    ):
        """Create a molecule object.

        If a mol_block is provided, the molecule will be initialized with that,
        otherwise the molecule will be initialized empty.

        Parameters
        ----------
        mol_block : str
            A mol block string.
        properties : dict[str, Any]
            A dictionary of properties.
        """

        self.except_on_invalid_molecule: bool = except_on_invalid_molecule
        self.properties: dict[str, Any] = properties
        self.mol_block: str = mol_block  # calls mol_block setter
        self.component_type: str = component_type

    @property
    def mol_block(self) -> str:
        """Returns the mol block string of the molecule.

        Returns
        -------
        str
            The mol block string of the molecule.
        """
        return self._mol_block

    @mol_block.setter
    def mol_block(self, mol_block: str):
        """Set the mol block string of the molecule.

        Parameters
        ----------
        mol_block : str
            A mol block string.
        """
        if not mol_block:
            self._mol_block = None
        else:
            self._from_mol_block(mol_block)

    @property
    def rd_mol(self) -> Mol:
        """Return the RDKit molecule object."""
        return MolFromMolBlock(
            self.mol_block,
        )

    @property
    def smiles(self) -> str:
        """Returns the SMILES string of the molecule using RDKit.

        Returns:
            str: SMILES string of the molecule.
        """
        try:
            return MolToSmiles(self.rd_mol)
        except Exception:
            return None

    @property
    def metadata(self) -> dict[str, Any]:
        """Returns the metadata of the molecule from the mol block string.

        Returns:
            dict[str, Any]: The metadata of the molecule.
        """
        return get_mol_block_metadata(self.mol_block)

    def _from_mol_block(self, mol_block: str, properties: dict[str, Any] = dict()) -> None:
        """Initialize the molecule object with a mol block string.

        Parameters
        ----------
        mol_block : str
            A mol block string.
        """
        self.properties.update(properties)
        self._mol_block = mol_block

        if self.except_on_invalid_molecule:
            try:
                assert self.rd_mol is not None
            except AssertionError:
                raise InvalidMoleculeError("mol_block is not a valid mol block string.")

    @classmethod
    def from_mol_block(cls, mol_block: str, properties: dict[str, Any] = {}) -> "Molecule":
        """Create a Molecule object from a mol block string.

        Parameters
        ----------
        mol_block : str
            A mol block string.

        Returns
        -------
        Molecule
            A Molecule object.
        """

        mol = cls()
        mol._from_mol_block(mol_block, properties)
        return mol

    def __eq__(self, __o: object) -> bool:
        """Returns True if the molecules are equal.

        Returns
        -------
        bool
            True if the molecules are equal.
        """
        if not isinstance(__o, Molecule):
            return False
        return self.mol_block == __o.mol_block

    def __str__(self) -> str:
        """Returns the smiles string of the molecule.

        Returns
        -------
        str
            The smiles string of the molecule.
        """
        return self.smiles

    def __repr__(self) -> str:
        return f"Molecule(smiles={self.smiles})"

    @classmethod
    def from_smiles(cls, smiles: str) -> "Molecule":
        """Create a Molecule object from a smiles string.

        Parameters
        ----------
        smiles : str
            A smiles string.

        Returns
        -------
        Molecule
            A Molecule object.
        """
        mol_block = MolToMolBlock(MolFromSmiles(smiles))
        mol = cls()
        mol._from_mol_block(mol_block)
        return mol


class Reactant(Molecule):
    def __init__(*args, **kwargs):
        super().__init__(**args, **kwargs, component_type="reactant")


class Product(Molecule):
    def __init__(*args, **kwargs):
        super().__init__(**args, **kwargs, component_type="product")


class Solvent(Molecule):
    def __init__(*args, **kwargs):
        super().__init__(**args, **kwargs, component_type="solvent")


class Catalyst(Molecule):
    def __init__(*args, **kwargs):
        super().__init__(**args, **kwargs, component_type="catalyst")



from __future__ import annotations

import logging
import re
from typing import Any, Callable, Iterator

from rdfreader.chem.mol import Molecule
from rdfreader.parse.utils import (
    CTF_COMPONENT_COUNT_FORMAT_STRING,
    CTF_DEFAULT_LETTER_TO_FIELD_MAPPING,
    CTF_RXNBLOCK_HEADER_FORMAT_STRING,
    _parse_block_header_line,
    get_whole_line_item,
    make_string_python_safe,
)

logger = logging.getLogger(__name__)


def get_rxn_block_metadata(
    rxn_block: str,
    header_format_string: str = CTF_RXNBLOCK_HEADER_FORMAT_STRING,
    header_field_mapping: dict[str, tuple[str, Callable, Any]] = CTF_DEFAULT_LETTER_TO_FIELD_MAPPING,
    reactant_product_count_format_string: str = CTF_COMPONENT_COUNT_FORMAT_STRING,  # noqa F401
) -> dict[str, Any]:
    """Get the metadata from a reaction block.

    Parameters
    ----------
    rxn_block : str
        A reaction block string.

    Returns
    -------
    dict
        A dictionary of metadata.
    """

    metadata = {}
    rxn_block_lines: list[str] = rxn_block.split("\n")
    metadata["reaction_name"] = get_whole_line_item(rxn_block_lines[1])
    metadata.update(_parse_block_header_line(rxn_block_lines[2], header_format_string, header_field_mapping))
    metadata["comment"] = get_whole_line_item(rxn_block_lines[3])
    metadata.update(
        _parse_block_header_line(
            rxn_block_lines[4],
            reactant_product_count_format_string,
            header_field_mapping,
        )
    )

    return metadata


class DatumParser:
    """Process dtype/datum pairs.

    The class processes the dtype string into a callable method name and then
    calls that method with the datum string to process the data.

    To use: instantiate the class and then call the class as a function with
    the dtype and datum strings.

    Example:
    >>> dp = DatumParser()
    >>> dp("$DATUM", "RXNVARIATION:STEPNO:SOLVENT:MOLSYMBOL")
    """

    def __init__(self, rxn_block: str, except_on_invalid_molecule: bool = True):
        self.rxn_block = rxn_block
        self.except_on_invalid_molecule = except_on_invalid_molecule

    def __iter__(self) -> Iterator[tuple[Any, int, str]]:
        """Iterate over the lines in the reaction block, identifying
        dtype/datum pairs and returning them as a tuple."""

        dtype_string_identifier = "$DTYPE "

        lines = self.rxn_block.split("\n")

        for line_idx, line in enumerate(lines):
            if line.startswith(dtype_string_identifier):
                dtype_string = line
                datum_string = ""
                # capture all the following lines until we find a line that
                # starts with $DTYPE again
                for datum_line in lines[line_idx + 1 :]:
                    if datum_line.startswith(dtype_string_identifier):
                        break

                    if datum_line.endswith("+"):
                        # if the line ends with a plus sign, it is a
                        # continuation of the previous line
                        datum_string += datum_line[:-1]
                    else:
                        datum_string += datum_line + "\n"

                yield self.parse_datum(dtype_string, datum_string)

    def __call__(self, *args, **kwargs):
        """Wraps parse_datum method."""
        return self.parse_datum(*args, **kwargs)

    def parse_datum(self, dtype: str, datum: str) -> tuple[str, str | Molecule]:
        """Process the $DATUM datum.

        Parameters
        ----------
        dtype : str
            The $DTYPE string from the reaction block.
        datum : str
            The corresponding datum string from the reaction block.

        Returns
        -------
        tuple[Any, str]
            A tuple of the processed datum and the parsed
            dtype string.
        """

        parsed_dtype = parse_dtype_string(dtype)
        datum = preprocess_datum_string(datum)

        # if the datum_string is a molblock:
        if detect_molblock_from_datum(datum):
            datum = preprocess_datum_molblock(datum)
            # try and infer the type from the dtype string, etc.)
            reagent_type = "other_reagent"
            for _reagent_type in ["catalyst", "solvent"]:
                if _reagent_type in parsed_dtype or _reagent_type.upper() in parsed_dtype:
                    reagent_type = _reagent_type

            datum = Molecule(
                mol_block=datum, component_type=reagent_type, except_on_invalid_molecule=self.except_on_invalid_molecule
            )
        else:
            datum = datum.strip()

        return parsed_dtype, datum


def detect_molblock_from_datum(datum: str) -> bool:
    """Tries to detect whether a datum string is a molblock."""

    if datum.startswith("$MFMT"):
        return True
    return False


def preprocess_datum_molblock(datum: str) -> str:
    """Strips the datum down to just the molblock."""

    return "\n".join(datum.splitlines()[1:])  # remove the $MFMT line


def preprocess_datum_string(datum: str) -> str:
    """Preprocess the datum string.

    Parameters
    ----------
    datum_str : str
        The datum string.

    Returns
    -------
    str
        The preprocessed datum string.
    """
    re_pattern = r"^\$DATUM\s?"
    datum = re.sub(re_pattern, "", datum)
    return datum


def parse_dtype_string(dtype_string: str) -> str:
    """Parse a $dtype line from a reaction block.

    Returns the contents of the dtype line, with the prefix "$DTYPE " removed.

    Parameters
    ----------
    dtype_string : str
        A string containing a $dtype line.

    Returns
    -------
    str
        The contents of the $dtype line.
    """

    dtype_string = dtype_string.strip()  # Remove leading and trailing whitespace.
    dtype_string = dtype_string.replace("$DTYPE ", "")  # Remove the $dtype tag.
    dtype_string = make_string_python_safe(dtype_string)
    return dtype_string


def validate_rxn_block(rxn_block: str) -> bool:
    """Validates a rxn block.

    Parameters
    ----------
    rxn_block : str
        The rxn block.

    Returns
    -------
    bool
        True if the rxn block is valid.
    """

    if rxn_block.startswith("$RXN"):
        return True
    else:
        return False


def mol_blocks_from_rxn_block(rxn_block: str, reactant_count: int, product_count: int) -> tuple[list[str], list[str]]:
    """Get the mol blocks corresponding to reactants and products from the rxn
    block.

    Params
    ------
    rxn_block : str
        A reaction block string.
    reactant_count : int
        The number of reactants in the reaction.
    product_count : int
        The number of products in the reaction.

    Returns
    -------
    reactants : list[str]
        The mol blocks corresponding to reactants.
    products : list[str]
        The mol blocks corresponding to products.
    """
    reactants: list[str] = []
    products: list[str] = []

    start_string = "$MOL\n"
    end_string = "M  END\n"

    # mol_blocks start with "$MOL" and end with "M  END"
    # so we can just split on these and get the reactants and products.
    # reactants are first, then products.
    mol_blocks: list[str] = rxn_block.split(start_string)

    # remove the first element as it is not a molblock
    mol_blocks.pop(0)

    if len(mol_blocks) > reactant_count + product_count:
        raise ValueError(
            "The number of mol blocks in the rxn block is greater than the number of reactants and products."  # noqa: E501
        )

    # if any element does not end with "M END", then remove all lines "M END".
    # This is important to capture the end of the last molblock correctly and
    # not include reaction data
    for ii, mol_block in enumerate(mol_blocks):
        if not mol_block.endswith(end_string):
            mol_blocks[ii] = mol_block.split(end_string)[0] + end_string  # add the end string back on

    # now we have the molblocks, we can split them into reactants and products
    reactants = mol_blocks[:reactant_count]
    products = mol_blocks[reactant_count:]

    return reactants, products



from rdfreader.chem.mol import Molecule


def mol_list_to_smiles(mol_list: list[Molecule]) -> str:
    """Convert a list of molecules to a SMILES string.

    Parameters
    ----------
    mol_list : list[Molecule]
        The list of molecules to convert to a SMILES string.

    Returns
    -------
    str
        The SMILES string.
    """

    return ".".join([mol.smiles for mol in mol_list if mol.smiles is not None])


def reaction_smiles(
    reactants: list[Molecule],
    products: list[Molecule],
    reagents: list[Molecule] = [],
) -> str:
    """Create a reaction smiles string from lists of product, reactant, and
    reagent molecules."""

    product_smiles = mol_list_to_smiles(products)
    reactant_smiles = mol_list_to_smiles(reactants)
    reagent_smiles = mol_list_to_smiles(reagents)

    return ">".join([reactant_smiles, reagent_smiles, product_smiles])



import datetime
import logging
import re
from typing import Any, Callable, Optional

logger = logging.getLogger(__name__)

# http://c4.cabrillo.edu/404/ctfile.pdf
CTF_DEFAULT_MOLBLOCK_HEADER_FORMAT_STRING: str = "IIPPPPPPPPMMDDYYHHmmddSSssssssssssEEEEEEEEEEEERRRRRR"
CTF_RXNBLOCK_HEADER_FORMAT_STRING: str = "IIIIIIPPPPPPPPPMMDDYYYYHHmmRRRRRRR"
SPRESI_RXNBLOCK_HEADER_FORMAT_STRING: str = "IIIIIIPPPPPPPPPPMMDDYYHHmmRRRRRRR"
CTF_COMPONENT_COUNT_FORMAT_STRING: str = "rrrppp"
# default mapping of letter to field. This is used to map the letter in the
# mol block header to the field in the metadata.
# key: letter, value: tuple containing the field name, the field type and a
# default value (if the field is empty). Field typle should be a castable data
# type.
# If the field is empty, the default value will be used.
CTF_DEFAULT_LETTER_TO_FIELD_MAPPING: dict[str, tuple[str, Callable, Any]] = {
    "I": ("user_initials", str, ""),
    "P": ("program_name", str, ""),
    "M": ("month", int, 0),
    "D": ("day", int, 0),
    "Y": ("year", int, 0),
    "H": ("hour", int, 0),
    "m": ("minute", int, 0),
    "d": ("dimensional_codes", str, ""),
    "S": ("scaling_factor_1", int, 0),
    "s": ("scaling_factor_2", float, 0.0),
    "E": ("energy", float, 0.0),
    "R": ("registry_number", str, ""),
    "r": ("reactant_count", int, 0),
    "p": ("product_count", int, 0),
}


def _default_line_item(cast_type: Callable = str, default: Any = None) -> Any:
    """Return a default value for a line item."""
    if default is not None:
        return default

    if cast_type in (int, float):
        return cast_type(0)
    if cast_type == str:
        return ""
    else:
        return default


def get_line_item(
    line: str,
    character_index: Optional[tuple[Optional[int], Optional[int]]] = None,
    cast_type: Callable = str,
    default: Any = None,
    catch_casting_exceptions: bool = False,
) -> Any:
    """Process a line from an rdf/mol file or block.

    If character_index is provided, then the characters within that slice will
    be returned.
    If cast_type is provided, then the characters will be cast to that type.
    If the line is empty a default will be provided. For numeric types (as set
    by cast_type), the default will be 0, for strings it will be an empty
    string. For other types, the default will be None. This can be overridden
    by providing a default.
    If catch_casting_exceptions is True, then any casting exceptions will be
    caught and the default will be returned.

    Strips whitespace and newlines.

    Parameters
    ----------
    line : str
        A line.
    character_index : tuple[int, int]
        A tuple of the start and end character index.

    Returns
    -------
    str
        The line item.
    """

    line = line[slice(*character_index)] if character_index else line
    line = line.strip("\n")  # remove newline
    line = line.strip()  # remove whitespace

    if not line:
        # return a default value if the line is empty
        return _default_line_item(cast_type, default)

    if cast_type:
        try:
            # attempt to cast the line to the specified type
            line = cast_type(line)
        except ValueError:
            if catch_casting_exceptions:
                # return a default value if the line cannot be cast
                return _default_line_item(cast_type, default)
            else:
                raise

    return line


def get_whole_line_item(line: str) -> str:
    """Return a data item that is from a whole line. Item must be a string.

    Parameters
    ----------
    line : str
        A line.

    Returns
    -------
    str
        The data item.
    """
    return get_line_item(line, (0, len(line)), str, "")


def parse_format_string(format_string) -> dict[str, tuple]:
    """Parses a CTF format string (http://c4.cabrillo.edu/404/ctfile.pdf) and
    returns a dictionary where the key is the letter in the format string and
    the value is a tuple of the start and end character index.

    Parameters
    ----------
    format_string : str
        A CTF format string.

    Returns
    -------
    dict[str, tuple]
        A dictionary of the letter in the format string and the start and end
        character index as a tuple.
    """
    format_string_dict = {}

    last_letter = format_string[0]
    current_letter_start_index = 0  # the index where the current letter started
    for ii, letter in enumerate(format_string):
        if letter != last_letter:
            # if the letter has changed add the previous letter to the
            # dictionary
            format_string_dict[last_letter] = (current_letter_start_index, ii)
            current_letter_start_index = ii

        last_letter = letter

    # add the last letter to the dictionary
    format_string_dict[last_letter] = (current_letter_start_index, ii + 1)

    return format_string_dict


def dict_elements_to_datetime(
    dd: dict[str, Any],
    date_time_key: str = "date_time",
    delete_initial_keys: bool = True,
    catch_datetime_exceptions: bool = True,
) -> dict[str, Any]:
    """Search a dictionary for datetime elements and adds a a new key to the
    dictionary with a datetime object.

    Searches for datetime keys called: "hour", "minute", "second", "day",
    "month", "year" and adds a new key called <date_time_key> with a datetime
    object.

    If delete_initial_keys is set, the original keys will be deleted.

    If the datetime is not parseable, None will be added to the dictionary.

    Parameters
    ----------
    dd : dict[str, Any]
        A dictionary.
    date_time_key : str
        The key to add to the dictionary.
    delete_initial_keys : bool
        If True, the keys "hour", "minute", "second", "day", "month", "year"
        will be deleted from the dictionary.
    catch_datetime_exceptions
        If True, parsing exceptions from datetime.datetime will be caught and
        None will be added to the dictionary.

    Returns
    -------
    dict[str, Any]
        A dictionary with the new key added.
    """

    date_time_args = {k: dd[k] for k in ["hour", "minute", "second", "day", "month", "year"] if k in dd}

    if not date_time_args:
        # if no datetime keys are found, return the dictionary as is
        return dd

    try:
        dd[date_time_key] = datetime.datetime(**date_time_args)
    except (ValueError, TypeError):
        if catch_datetime_exceptions:
            logger.warning(f"Could not parse datetime from {dd}")
            dd[date_time_key] = None
        else:
            raise

    if delete_initial_keys:
        [dd.pop(key, None) for key in ["hour", "minute", "second", "day", "month", "year"]]

    return dd


def _parse_block_header_line(
    header_line: str,
    header_format_string: str,
    header_field_mapping: dict[str, tuple[str, Callable, Any]],
) -> dict[str, Any]:
    """Parse the header line of a rxn or mol block.

    Parameters
    ----------
    header_line : list[str]
        A single line from the header.
    header_format_string : str
        See get_mol_block_metadata for more information.
    header_field_mapping : dict[str, tuple(str, Callable, Any)]
        See get_mol_block_metadata for more information.


    Returns
    -------
    dict
        A dictionary of metadata.
    """

    format_string_dict = parse_format_string(header_format_string)

    metadata = {}
    for letter, character_index in format_string_dict.items():

        if letter in header_field_mapping:
            field_name = header_field_mapping[letter][0]
            data_type = header_field_mapping[letter][1]
            try:
                default_value = header_field_mapping[letter][2]
            except IndexError:
                default_value = None
            metadata[field_name] = get_line_item(header_line, character_index, data_type, default_value)
        else:
            raise ValueError(f"The letter {letter} does not appear in the format field mapping.")  # noqa: E501

    metadata = dict_elements_to_datetime(metadata)

    return metadata


def make_string_python_safe(string: str) -> str:
    """Remove/replace characters that are not allowed in python
    variable/function names.

    Parameters
    ----------
    string : str
        A string.

    Returns
    -------
    str
        The python safe string.
    """

    if string is None:
        return None

    string = string.strip()
    string = string.strip("\n")
    string = string.strip("\r")

    if string == "":
        return None

    # remove existing underscores
    # string = string.replace("_", "")

    # add a leading underscore if string starts with a number
    if string[0].isdigit():
        string = f"_{string}"

    # replace all non-alphanumeric characters with an underscore
    string = re.sub(r"[^a-zA-Z0-9_]", "_", string)

    # lower case the string
    string = string.lower()

    # remove trailing underscores
    string = string.rstrip("_")

    return string


def parse_yield(yield_string: str) -> float:
    """Attempts to parse the yield string into a float.

    If muliple numbers are detected, will return the average.

    If no numbers are detected, will return None.

    Parameters
    ----------
    yield_string : str

    Returns
    -------
    float
        Yield as a float.
    """
    re_patterns = [
        r"^([0-9]+\.?[0-9]?)$",  # matches a single int or float
        r"^([0-9]+\.?[0-9]?)\s{0,}[-,;:]{0,}\s{0,}([0-9]+\.?[0-9]?)$",  # matches two ints or floats separated by a dash, comma, semicolon, colon or space  # noqa: E501
    ]

    for re_pattern in re_patterns:
        match = re.search(re_pattern, yield_string)
        if match:
            # get the match groups as a list
            # (first match group is the whole string)
            match_groups = match.groups()
            # convert the match groups to floats
            match_groups = [float(group) for group in match_groups]
            # average the match groups
            return sum(match_groups) / len(match_groups)

    # if we get here, then we didn't find a match
    logger.warning(f"Could not parse yield from '{yield_string}'. Returning None.")

    return None



# typing
from io import TextIOWrapper
from pathlib import Path

from rdfreader.chem.reaction import Reaction
from rdfreader.parse.utils import CTF_RXNBLOCK_HEADER_FORMAT_STRING


def parse_rdf_reg_num(line: str):
    return line.replace("$RFMT $RIREG ", "").strip()


class RDFParser:
    _header_retrieved: bool = False
    lineno: int = 1
    rdf_metadata: dict[str, str] = {}

    def __init__(
        self,
        f: TextIOWrapper,
        header_format_string: str = CTF_RXNBLOCK_HEADER_FORMAT_STRING,
        except_on_invalid_molecule: bool = True,
        except_on_invalid_reaction: bool = True,
        parse_conditions: bool = True,
    ):
        """
        Parameters
        ----------
        f : TextIOWrapper
            The file to parse.
        header_format_string : str, optional
            The format string to use to parse the header of each rxn block.
        except_on_invalid_molecule : bool, optional
            If True, raise an exception if a molecule is invalid.
        except_on_invalid_reaction : bool, optional
            If True, raise an exception if a reaction is invalid.
        rdf_file_name : str, optional
            The name of the rdf file.
        parse_conditions : bool, optional
            Whether to parse the conditions of the reaction.
        """

        self.f = f
        self.header_format_string = header_format_string
        self.except_on_invalid_molecule = except_on_invalid_molecule
        self.except_on_invalid_reaction = except_on_invalid_reaction
        self.rdf_file_name = Path(f.name).name
        self.parse_conditions = parse_conditions

    def __iter__(self):
        return self

    def __next__(self):
        return self.next_reaction()

    def next_reaction(self):
        """Returns the next rxn block from a rdf file.

        Parameters
        ----------
        f : TextIOWrapper
            The file to parse.

        Returns
        -------
        tuple[str, str, int]
            The rxn block, the rxn id, and the line number of the start of the
            rxn block.
        """

        # get the next rxn block
        reaction = None
        rxn_block, rxn_id, start_lineno = self._next_rxn_block()
        try:
            reaction = Reaction(
                rxn_block=rxn_block,
                id=rxn_id,
                rdf_metadata=self.rdf_metadata,
                header_format_string=self.header_format_string,
                except_on_invalid_molecule=self.except_on_invalid_molecule,
                lineno=start_lineno,
                rdf_file=self.rdf_file_name,
            )

        except Exception as e:
            if self.except_on_invalid_reaction:
                raise e

        return reaction

    def _header(self):
        """Parse the header of a RDF file.

        Parameters
        ----------
        f : TextIOWrapper
            The file to parse.

        Returns
        -------
        dict[str, str]
            The version and date of the RDF file. These are just treated as
            strings as they are typically ignored and the structure of the
            datetime field is not defined in the specification.
        """
        if not self._header_retrieved:
            self._header_retrieved = True
            version = self._readline()[8:].strip()
            date_stamp = self._readline()[6:].strip()
            self.rdf_metadata = {"version": version, "date_stamp": date_stamp}

    def _next_rxn_block(self) -> tuple[str, str, int]:
        """Returns the next rxn block from a rdf file. If the end.

        Returns
        -------
        tuple[str, str, int]
            The rxn block, the rxn id, and the line number of the start of the
            rxn block.
        """

        self._header()
        start_lineno: int = self.lineno
        line: str = self._readline()
        if line == "":
            raise StopIteration
        # parse the rxn block deliminators
        if line.startswith("$RFMT"):
            # capture the reg number
            reg_no: str = parse_rdf_reg_num(line)
            line: str = self._readline()
        else:
            #  there is a problem with the file format, raise an exception
            raise Exception(f"Invalid RDF file format. Expected $RFMT, got {line} " f"at line {self.lineno}")

        # capture the rxn block
        rxn_block: str = ""
        f_last_pos = self.f.tell()  # ensure f_last_pos is defined
        while not line.startswith("$RFMT") and not line == "":
            f_last_pos = self.f.tell()
            rxn_block += line
            line: str = self._readline()

        # send the file pointer back one line so it is at the start of the
        # next rxn block
        self.f.seek(f_last_pos)
        self.lineno -= 1

        return rxn_block, reg_no, start_lineno

    def _readline(self):
        """Wraps f.read and increments the line number."""
        self.lineno += 1
        return self.f.readline()



import logging
from typing import Any, Optional

from rdkit.Chem.rdChemReactions import ChemicalReaction, ReactionFromSmarts

from rdfreader.chem.mol import Molecule
from rdfreader.chem.utils import reaction_smiles
from rdfreader.exceptions import InvalidReactionError
from rdfreader.parse.rxnblock import DatumParser, get_rxn_block_metadata, mol_blocks_from_rxn_block, validate_rxn_block
from rdfreader.parse.utils import CTF_RXNBLOCK_HEADER_FORMAT_STRING

logger = logging.getLogger(__name__)


class Reaction:
    def __init__(
        self,
        rxn_block: str = None,
        id: str = None,
        rdf_metadata: Optional[dict[str, Any]] = None,  # metadata from the rdf file the block was contained in,
        except_on_invalid_molecule: bool = True,
        header_format_string: str = CTF_RXNBLOCK_HEADER_FORMAT_STRING,
        lineno: Optional[int] = None,
        rdf_file: Optional[str] = None,
    ):
        """Create a reaction object.

        If a reaction block is provided, the reaction will be
        initialized with that, otherwise the reaction will be
        initialized empty.
        """

        self.rxn_block = rxn_block
        self.rdf_metadata = rdf_metadata
        self.id = id
        self.lineno = lineno
        self.rdf_file = rdf_file

        self.products: list[Molecule] = list()
        self.reactants: list[Molecule] = list()
        self.catalysts: list[Molecule] = list()
        self.solvents: list[Molecule] = list()
        self.other_reagents: list[Molecule] = list()

        self.properties: dict[str, str] = dict()
        self.metadata: dict[str, Any] = dict()

        if self.rxn_block is not None:
            self._from_rxn_block(
                header_format_string=header_format_string,
                except_on_invalid_molecule=except_on_invalid_molecule,
            )

    def __str__(self) -> str:
        return f"Reaction({self.id}, {self.smiles})"

    def _from_rxn_block(
        self,
        header_format_string: str = CTF_RXNBLOCK_HEADER_FORMAT_STRING,
        except_on_invalid_molecule: bool = True,
    ) -> None:
        """Initialize the reaction object from a reaction block.

        Parameters
        ----------
        rdf_path : Path
            A path to a rdf file.
        """

        if not validate_rxn_block(self.rxn_block):
            raise ValueError("Reaction block is invalid.")

        # get the headers from the rxn block.
        self.metadata.update(get_rxn_block_metadata(self.rxn_block, header_format_string=header_format_string))

        # get the reactants and product mol_blocks, create a molecule object
        # for each, and add it to the reaction.
        reactant_mol_blocks, product_mol_blocks = mol_blocks_from_rxn_block(
            self.rxn_block,
            self.metadata["reactant_count"],
            self.metadata["product_count"],
        )
        for mol_block in reactant_mol_blocks:
            self.reactants.append(
                Molecule(
                    mol_block,
                    except_on_invalid_molecule=except_on_invalid_molecule,
                )
            )

        for mol_block in product_mol_blocks:
            self.products.append(
                Molecule(
                    mol_block,
                    except_on_invalid_molecule=except_on_invalid_molecule,
                )
            )

        # use functions in rdfreader/parse/rdf.py to pull out dtype/datum
        # pairs from the rxn block
        datum_parser = DatumParser(
            self.rxn_block,
        )

        for dtype, datum in datum_parser:
            if not isinstance(datum, Molecule):
                self.properties[dtype] = datum
            else:
                # add to the appropriate molecule list
                getattr(self, f"{datum.component_type}s").append(datum)

        if self.rd_rxn is None:
            raise ValueError("Invalid reaction: couldn't parse in rdkit.")

    @property
    def smiles(self) -> str:
        return reaction_smiles(
            self.reactants,
            self.products,
            self.reagents,
        )

    @property
    def smiles_no_reagents(self) -> str:
        return reaction_smiles(
            self.reactants,
            self.products,
        )

    @property
    def rd_rxn(self) -> ChemicalReaction:
        try:
            return ReactionFromSmarts(self.smiles, useSmiles=True)
        except ValueError as e:
            raise InvalidReactionError(f"Invalid reaction: {e}") from e

    @property
    def reagents(self) -> list[Molecule]:
        """Return a single list of all reagents.

        Returns
        -------
        list[Molecule]
            A list of all reagents.
        """
        return self.catalysts + self.solvents + self.other_reagents



class InvalidMoleculeError(Exception):
    """Raised when the mol block string is not a valid mol block string."""


class InvalidReactionError(Exception):
    """Raised when the reaction block string is not a valid reaction block
    string."""



import logging
from typing import Any, Callable

from rdfreader.parse.utils import (
    CTF_DEFAULT_LETTER_TO_FIELD_MAPPING,
    CTF_DEFAULT_MOLBLOCK_HEADER_FORMAT_STRING,
    _parse_block_header_line,
    get_line_item,
    get_whole_line_item,
)

logger = logging.getLogger(__name__)


def get_mol_block_metadata(
    mol_block: str,
    header_format_string: str = CTF_DEFAULT_MOLBLOCK_HEADER_FORMAT_STRING,
    header_field_mapping: dict[str, tuple[str, Callable, Any]] = CTF_DEFAULT_LETTER_TO_FIELD_MAPPING,
) -> dict[str, Any]:
    """Extract metadata from a mol block string.

    Parameters
    ----------
    mol_block : str
        A mol block string.
    header_format_string : str
        The format string for the line 2 of the mol_block header. Default
        value is according to the mol_block spec at
        http://c4.cabrillo.edu/404/ctfile.pdf. The letters must appear in the
        spec, but the order can be changed.
    header_field_mapping : dict[str, tuple(str, Callable, Any)]
        A dictionary which maps a letter in the header format string to a
        field name. The field name is the key and the letter is the value. The
        default mapping is according to the mol_block spec at
        http://c4.cabrillo.edu/404/ctfile.pdf.
        key: letter, value: tuple containing the field name, the field type
        and a default value (if the field is empty). Field typle should be a
        castable data type.
        If the field is empty, the default value will be used.

    Returns
    -------
    dict
        A dictionary of metadata.
    """

    metadata = {}
    mol_block_lines = mol_block.split("\n")
    metadata["molecule_name"] = get_whole_line_item(mol_block_lines[0])
    metadata.update(_parse_block_header_line(mol_block_lines[1], header_format_string, header_field_mapping))
    metadata["comment"] = get_whole_line_item(mol_block_lines[2])
    large_regno: str = _parse_large_regno(mol_block_lines)
    if large_regno is not None:
        # overwrite the registry number from the header with the large
        # registry number if it is present
        metadata["registry_number"] = large_regno

    return metadata


def _parse_large_regno(mol_block_lines: list[str]) -> str:
    """Searches the molblock a line beginning with M REG and returns the value
    of it if present.

    Parameters
    ----------
    mol_block_lines : list[str]
        A molblock string that has been split into a list of lines.

    Returns
    -------
    Optional[str]
        The registry number. If None is returned, the registry number is not
        present.
    """

    for line in mol_block_lines:
        if line.startswith("M  REG "):
            return get_line_item(line, [7, len(line)])

    return None
